'******************************************************************************
' C Function Library in Spin
' Author: Dave Hein
' Copyright (c) 2010, 2011, 2012
' See end of file for terms of use.
'******************************************************************************
'******************************************************************************
' Revison History
' v1.0 - 4/2/2010 First official release
'******************************************************************************
{{ 
  This object provides file I/O routines that use the standard C calling
  interface.  It is built upon FSRW 2.6 with modifications to support
  subdirectories and multiple open files with a single object.  Each open file
  uses a file descriptor that contains a copy of the 10 FSRW state variables.
  There is an additional state variable that points to a dedicated sector
  buffer that is 512 bytes in size.  The total size of a file descriptor is
  556 bytes.

  The directory functions opendir and readdir use the same file descriptor as
  normal files.  readdir returns the 32-byte directory entry for a file plus
  the file name the 8.3 format.  The stat function also returns the 32-byte
  directory entry, but without the formatted file name at the end.

  cfileio uses the cmalloc.spin memory manager to allocate buffer space for
  the file descriptors.  The programmer could provide his own memory manager
  as long as it provides malloc and free functions.
}}

OBJ
  mem   : "cmalloc"     ' Memory manager used to allocate file descriptors
  sdspi : "csafe_spi"   ' SD driver that provides block-level access.

CON
  ' Error codes used for errno
  ENOENT        = 304
  EACCES        = 306
  ENOMEM        = 307
  EEXIST        = 309
  EINVAL        = 310
  EMFILE        = 311
  EIO           = 312
  ENOTDIR       = 313
  EISDIR        = 314
  EROFS         = 315
  ENOTEMPTY     = 317
  ENAMETOOLONG  = 318
  ENOSPC        = 324

  ' Values for the fseek origin parameter
  SEEK_SET      = 0
  SEEK_CUR      = 1
  SEEK_END      = 2

  ' Value of the newline character
  NEW_LINE      = 10

  ' Size of the file descriptor in bytes and longs
  HANDLE_SIZE   = 44
  HANDLE_LONGS  = 11

DAT
  errno         long 0          ' Error number from last I/O operation
  currentwd     long 0[25]      ' Current Working Directory
  readdir_block long 0[12]      ' Directory entry plus d_name

PUB mount(basepin)
  return mount_explicit(basepin, basepin + 1, basepin + 2, basepin + 3)

PUB mount_explicit(DO, CLK, DI, CS)
  result := hmount_explicit(DO, CLK, DI, CS)
  bytemove(@currentwd, string("/"), 2)

PUB fopen(fname, mode) | handle, modechar, pathstr[20], filestr[8]
{{
  This routine opens a file with the name pointed to by "fname".  The mode is
  determined by the first character of the string pointed to by "mode".  The
  first character must be "r", "w" or "a" for read, write or append.
  It returns a pointer to the file descriptor if successful, or a NULL
  value if not successful.
}}
  modechar := byte[mode]
  if modechar <> "r" and modechar <> "w" and modechar <> "a"
    errno := EINVAL
    return
  if ProcessPath(fname, @pathstr, @filestr)
    return
  ifnot (handle := mem.malloc(HANDLE_SIZE + SECTORSIZE))
    errno := ENOMEM
    return
  longfill(@fclust, 0, 10)
  pbuf := handle + HANDLE_SIZE
  result := \popen(@filestr, modechar)
  if result
    errno := -result
    mem.free(handle)
    return 0
  longmove(handle, @fclust, HANDLE_LONGS)
  return handle

PUB fclose(handle)
{{
  This routine closes an opened file.
}}
  longmove(@fclust, handle, HANDLE_LONGS)
  result := \pclose
  mem.free(handle)

PUB fread(buffer, elem_size, num, handle)
{{
  This routine reads a block of data from a file.  The parameter "handle"
  points to the file descriptor used to access the file.  This routine returns
  the number of bytes read, or a 0 no data is read due to an error condition or
  end of file.
}}
  longmove(@fclust, handle, HANDLE_LONGS)
  result := \pread(buffer, elem_size * num)
  longmove(handle, @fclust, HANDLE_LONGS)
  if result < 0
    errno := -result
    result := 0
  
PUB fgetc(handle)
{{
  This routine returns a single character from a file.  It will return a
  negative number if there is an error condition or end of file.
}}
  longmove(@fclust, handle, HANDLE_LONGS)
  result := \pgetc
  longmove(handle, @fclust, HANDLE_LONGS)
  if result < 0
    errno := -result

PUB fgets(str, num, handle) | i, value
{{
  This routine returns a string from a file.  It will read up to num-1 bytes,
  or until a newline character is encountered.  The newline character will be
  included if read from the file.  If will return the value of the string
  pointer, or NULL if an error condition or end of file is encountered.
}}
  result := str
  longmove(@fclust, handle, HANDLE_LONGS)
  repeat i from 0 to num - 2
    value := \pgetc
    if (value < 0)
      if (i == 0)
        longmove(handle, @fclust, HANDLE_LONGS)
        return 0
      quit
    byte[str++] := value
    if value == NEW_LINE
      quit
  byte[str] := 0
  longmove(handle, @fclust, HANDLE_LONGS)

PUB fwrite(buffer, elem_size, num, handle)
{{
  This routine writes a block of data to a file.  The parameter "handle" points
  to the file descriptor used to access the file.  This routine returns the
  number of bytes written, or a 0 if an error condition is encountered.
}}
  longmove(@fclust, handle, HANDLE_LONGS)
  result := \pwrite(buffer, elem_size * num)
  longmove(handle, @fclust, HANDLE_LONGS)
  if result < 0
    errno := -result
    result := 0

PUB fputc(value, handle)
{{
  This routine writes a single byte to a file.
}}
  longmove(@fclust, handle, HANDLE_LONGS)
  result := \pputc(value)
  longmove(handle, @fclust, HANDLE_LONGS)
  if result < 0
    errno := -result

PUB fputs(str, num, handle)
{{
  This routine writes a string to a file.
}}
  result := fwrite(str, 1, strsize(str), handle)

PUB remove(fname) | pathstr[20], filestr[8]
{{
  This routine delete a file.  It will return 0 on success, or a negative
  number on failure.
}}
  if (result := ProcessPath(fname, @pathstr, @filestr))
    return
  pbuf := @buf
  direntry := 0
  result := \popen(@filestr, "d")
  if result
    errno := -result
    
PUB rename(fname1, fname2) | ptr
{{
  This routine rename a file.  It will return 0 on success, or a negative
  number on failure.
}}
  pbuf := @buf
  direntry := 0
  result := \popen(fname2, "r")
  ifnot result
    \pclose
    return -1
  result := \popen(fname1, "r")
  if result
    return
  \pclose
  ptr := readbytec(direntryrd)
  ConvertName(fname2, ptr)
  dirty := 1
  flushifdirty
  result := 0

PUB opendir(path) | handle, value, pathstr[20], filestr[8]
{{
  This routine opens a directory for reading with the readdir routine.  It
  return a pointer to the directory descriptor if successfule, or a NULL on
  failure.
}}
  ResolvePath(path, @pathstr)
  if (result := changedir(@pathstr))
    errno := -result
    return 0
  ifnot (handle := mem.malloc(HANDLE_SIZE + SECTORSIZE))
    return 0
  longfill(@fclust, 0, 10)
  pbuf := handle + HANDLE_SIZE
  value := \popendir
  longmove(handle, @fclust, HANDLE_LONGS)
  if value
    mem.free(handle)
    return 0
  return handle

PUB readdir(handle) | retval, attrib, ptr1, ptr2, val 
{{
  This routine reads the 32-byte directory entry for the next file in the
  directory.  The file name is converted to the 8.3 format, and appended after
  the 32-byte entry.  On success, a the address of the readdir_block is
  returned. A NULL pointer is return if there are no more directory entries
  or if an error contidion is encountered.
}}
  longmove(@fclust, handle, HANDLE_LONGS)
  ptr2 := @readdir_block
  repeat
    retval := \pread(ptr2, 32) ' Read a directory entry
    if retval < 0
      longmove(handle, @fclust, HANDLE_LONGS)
      return
    ifnot byte[ptr2] ' Check for the end of the directory
      longmove(handle, @fclust, HANDLE_LONGS)
      return
    attrib := byte[ptr2 + 11] ' Get the attribute byte
    if (byte[ptr2] <> $e5) and ((attrib & $08) == 0) ' Exit loop if a valid file or directory entry
      quit
  ptr1 := @readdir_block + 32
  ' Get up to eight file name characters
  repeat 8
    val := byte[ptr2++]
    if val =< 32 or val > 126
      quit
    byte[ptr1++] := val
  ' Check for an extension  
  ptr2 := @readdir_block[2]
  val := byte[ptr2]
  if val > 32 and val =< 128
    byte[ptr1++] := "."
    ' Get up to 3 extension characters
    repeat 3
      val := byte[ptr2++]
      if val =< 32 or val > 126
        quit
      byte[ptr1++] := val
  byte[ptr1] := 0 ' NULL terminate the file name string
  longmove(handle, @fclust, HANDLE_LONGS)
  return @readdir_block 

PUB closedir(handle)
{{
  This routine closes an opened directory.
}}
  longmove(@fclust, handle, HANDLE_LONGS)
  result := \pclose
  mem.free(handle)

PUB ftell(handle)
{{
  This routine returns the current file position.
}}
  longmove(@fclust, handle, HANDLE_LONGS)
  result := floc + bufat - bufend

PUB fseek(handle, position, origin)
{{
  This routine moves to the location in the file determined by
  the "position" and the "origin".  For origin values of 0, 1 or 2
  the base position is the start of the file, the current position
  or the end of the file respectively.
}}
  longmove(@fclust, handle, HANDLE_LONGS)
  if origin == SEEK_CUR
    position += floc + bufat - bufend
  elseif origin == SEEK_END
    position += filesize
  result := \seek(position)
  longmove(handle, @fclust, HANDLE_LONGS)

PUB fflush(handle)
  longmove(@fclust, handle, HANDLE_LONGS)
  result := \pflushbuf(bufat, 1)
  longmove(handle, @fclust, HANDLE_LONGS)

PUB chdir(path) | abspath[25], cwdlen, abslen, fname
  ' Check if current directory
  if byte[path] == 0 OR strcomp(path, string("."))
    return
  ResolvePath(path, @abspath)
  result := changedir(@abspath)
  ifnot result
    strcpy(@currentwd, @abspath)
  else
    errno := -result

PUB getcwd(str, num)
  result := str
  strncpy(str, @currentwd, num)

PUB mkdir(fname, mode) | pathstr[20], filestr[8]
  if (result := ProcessPath(fname, @pathstr, @filestr))
    return
  pbuf := @buf
  direntry := 0
  result := \popen(@filestr, "r")
  ifnot result
    \pclose
    errno := EEXIST
    return -1
  if result <> -ENOENT
    errno := -result
    return
  result := \popen(@filestr, "w")
  if result
    errno := -result
    return
  byte[readbytec(direntry)+11] := %110000 ' (%110001 & mode) ^ 1
  dirty := 1
  longfill(@filestr, 0, 8)
  \pwrite(@filestr, 32)
  'floc := -32 ' Make floc + bufat = 0, so directory has zero size
  \pclose
  flushifdirty

PUB rmdir(fname) | handle, pathstr[20], filestr[8]
  handle := opendir(fname)
  ifnot handle
    return -1
  result := readdir(handle)
  closedir(handle)
  if result
    errno := ENOTEMPTY
    return -1
  ProcessPath(fname, @pathstr, @filestr)
  pbuf := @buf
  direntry := 0
  result := \popen(@filestr, "D")
  if result
    errno := -result

PUB chmod(fname, modebits) | pathstr[20], filestr[8]
  if (result := ProcessPath(fname, @pathstr, @filestr))
    return
  pbuf := @buf
  direntry := 0
  if (result := \popen(@filestr, "r"))
    errno := -result
    return
  byte[readbytec(direntryrd)+11] := modebits & $21
  dirty := 1
  flushifdirty

PUB stat(fname, statbuf) | pathstr[20], filestr[8]
  if (result := ProcessPath(fname, @pathstr, @filestr))
    return
  pbuf := @buf
  direntry := 0
  if (result := \popen(@filestr, "r"))
    errno := -result
    return
  bytemove(statbuf, readbytec(direntryrd), 32)

PUB getfilesize(handle)
  result := long[handle][1]

PUB strerror(errnum)
  case errnum
    ENOENT       : result := string("No such file or directory")
    EACCES       : result := string("Permission denied")
    ENOMEM       : result := string("Not enough memory")
    EEXIST       : result := string("File exists")
    EINVAL       : result := string("Invalid argument")
    EMFILE       : result := string("Too many open files")
    EIO          : result := string("I/O error")
    ENOTDIR      : result := string("Not a directory")
    EISDIR       : result := string("Is a directory")
    EROFS        : result := string("Read only file system")
    ENOTEMPTY    : result := string("Directory not empty")
    ENAMETOOLONG : result := string("Name too long")
    ENOSPC       : result := string("No space on device")
    other        : result := string("Unknown error")

PUB get_errno
  result := errno

'=============================================================================
' Private directory routines
'=============================================================================

PRI ResolvePath(fname, path) | ptr, ptr1, pathLen
  if strcomp(fname, string("."))
    fname++
  elseif (not strncmp(fname, string("./"), 2) or not strncmp(fname, string(".\"), 2))
    fname += 2
  if (byte[fname][0] == "/" or byte[fname][0] == "\")
    strcpy(path, fname)
  elseif (byte[fname][0] == 0)
    strcpy(path, @currentwd)
  else
    strcpy(path, @currentwd)
    pathLen := strsize(path)
    if (byte[path][pathLen - 1] <> "/" and byte[path][pathLen - 1] <> "\")
      strcat(path, string("/"))
    strcat(path, fname)
  ' Whack DOS paths to UNIX paths
  ptr := path
  repeat while (byte[ptr])
    if (byte[ptr] == "\")
      byte[ptr] := "/"
    ptr++
  ' Process ..
  ptr := path
  repeat while (byte[ptr])
    if (not strncmp(ptr, string("/.."), 3) and (byte[ptr][3] == 0 or byte[ptr][3] == "/"))
      if (ptr == path)
        bytemove(ptr, ptr + 3, strsize(ptr))
      else
        ptr1 := ptr - 1
        repeat while (ptr1 <> path)
          if (byte[ptr1] == "/")
            quit
          ptr1--
        bytemove(ptr1, ptr + 3, strsize(ptr))
        ptr := ptr1
    else
      ptr++
      repeat while (byte[ptr])
        if (byte[ptr] == "/")
          quit
        ptr++
  if (byte[path][0] == 0)
    strcpy(path, string("/"))

PRI SplitPathFile(pathstr, filestr) | fname, len, len1, len2
  len2 := 1
  len1 := len := strsize(pathstr)
  ' Check for zero-length string
  ifnot len
    byte[pathstr] := "/"
    byte[pathstr][1] := 0
    byte[filestr] := 0
    return
  ' Check for root-directory only
  if len == 1 and byte[pathstr] == "/"
    byte[filestr] := 0
    return
  ' Search backwards for a "/"
  repeat while --len > 0
    if byte[pathstr][len] == "/"
      byte[pathstr][len] := 0
      bytemove(filestr, pathstr+len+1, len2)
      return
    len2++
  if byte[pathstr] == "/"
    bytemove(filestr, pathstr+1, len2)
  else
    bytemove(filestr, pathstr, len2+1)
    byte[pathstr] := "/"
  byte[pathstr][1] := 0

PRI ProcessPath(fname, pathstr, filestr)
  ResolvePath(fname, pathstr)
  SplitPathFile(pathstr, filestr)
  if (result := changedir(pathstr))
    errno := -result

'******************************************************************************  
' Private string routines
'******************************************************************************  
PRI FindChar(str, value)
  repeat while (byte[str])
    if (byte[str] == value)
      quit
    str++
  return str

PRI strcpy(dst, src)
  bytemove(dst, src, strsize(src) + 1)

PRI strncpy(dst, src, num) | slen
  if (--num => 0)
    slen := strsize(src) <# num
    bytemove(dst, src, slen)
    byte[dst][slen] := 0

PRI strncmp(ptr1, ptr2, num)
  repeat num
    if byte[ptr1++] <> byte[ptr2++]
      return 1

PRI strcat(dst, src)
  bytemove(dst + strsize(dst), src, strsize(src)+1)

{{
'   fsrw 2.6 Copyright 2009  Tomas Rokicki and Jonathan Dummer
'
'   See end of file for terms of use.
'
'   This object provides FAT16/32 file read/write access on a block device.
'   Only one file open at a time.  Open modes are 'r' (read), 'a' (append),
'   'w' (write), and 'd' (delete).  Only the root directory is supported.
'   No long filenames are supported.  We also support traversing the
'   root directory.
'
'   In general, negative return values are errors; positive return
'   values are success.  Other than -1 on popen when the file does not
'   exist, all negative return values will be "aborted" rather than
'   returned.
'
}}
'
'   Constants describing FAT volumes.
'
con
   SECTORSIZE  = 512
   SECTORSHIFT = 9
   DIRSIZE     = 32
   DIRSHIFT    = 5

dat
'
'   Variables concerning the open file.
'
   fclust        long 0 ' the current cluster number
   filesize      long 0 ' the total current size of the file
   floc          long 0 ' the seek position of the file
   frem          long 0 ' how many bytes remain in this cluster from this file
   bufat         long 0 ' where in the buffer our current character is
   bufend        long 0 ' the last valid character (read) or free position (write)
   direntry      long 0 ' the byte address of the directory entry (if open for write)
   writelink     long 0 ' the byte offset of the disk location to store a new cluster
   fatptr        long 0 ' the byte address of the most recently written fat entry
   firstcluster  long 0 ' the first cluster of this file
   pbuf          long 0 ' pointer to the sector buffer

'
'   Variables used when mounting to describe the FAT layout of the card
'
   filesystem    long 0 ' 0 = unmounted, 1 = fat16, 2 = fat32
   currdirstart  long 0 ' the byte address of the start of the current directory
   currdirend    long 0 ' the byte immediately following the current directory.
   dataregion    long 0 ' the start of the data region, offset by two sectors
   clustershift  long 0 ' log base 2 of blocks per cluster
   clustersize   long 0 ' total size of cluster in bytes
   fat1          long 0 ' the block address of the fat1 space
   totclusters   long 0 ' how many clusters in the volume
   sectorsperfat long 0 ' how many sectors per fat
   endofchain    long 0 ' end of chain marker (with a 0 at the end)
   pdate         long 0 ' current date
   lastread      long 0 ' the block address of the buf2 contents
   dirty         long 0 ' nonzero if buf2 is dirty
   rootdirstart  long 0 ' the byte address of the start of the root directory
   rootdirend    long 0 ' the byte immediately following the root directory
   direntryrd    long 0 ' the byte address of the directory entry (if open for read)
   currdirstr    long 0[20]          ' Current FSRW directory
   buf2          byte 0[SECTORSIZE]  ' main metadata buffer
   buf           byte 0[SECTORSIZE]  ' data buffer used for general file operations
   padname       byte 0[12]          ' filename buffer

pri release
'
'   This is just a pass-through function to allow the block layer
'   to tristate the I/O pins to the card.
'
   sdspi.release

pri writeblock2(n, b)
'
'   On metadata writes, if we are updating the FAT region, also update
'   the second FAT region.
'
   sdspi.writeblock(n, b)
   if (n => fat1)
     if (n < fat1 + sectorsperfat)
       sdspi.writeblock(n+sectorsperfat, b)

pri flushifdirty
'
'   If the metadata block is dirty, write it out.
'
   if (dirty)
      writeblock2(lastread, @buf2)
      dirty := 0

pri readblockc(n)
'
'   Read a block into the metadata buffer, if that block is not already
'   there.
'
   if (n <> lastread)
      flushifdirty
      sdspi.readblock(n, @buf2)
      lastread := n

pri brword(b)
'
'   Read a byte-reversed word from a (possibly odd) address.
'
   return (byte[b]) + ((byte[b][1]) << 8)

pri brlong(b)
'
'   Read a byte-reversed long from a (possibly odd) address.
'
   return brword(b) + (brword(b+2) << 16)

pri brclust(b)
'
'   Read a cluster entry.
'
  if (filesystem == 1)
    return brword(b)
  else
    return brlong(b)

pri brwword(w, v)
'
'   Write a byte-reversed word to a (possibly odd) address, and
'   mark the metadata buffer as dirty.
'
   byte[w++] := v
   byte[w] := v >> 8
   dirty := 1

pri brwlong(w, v)
'
'   Write a byte-reversed long to a (possibly odd) address, and
'   mark the metadata buffer as dirty.
'
   brwword(w, v)
   brwword(w+2, v >> 16)

pri brwclust(w, v)
'
'   Write a cluster entry.
   if (filesystem == 1)
     brwword(w, v)
   else
     brwlong(w, v)
'
'   This may do more complicated stuff later.
'
pri unmount
  pclose
  'sdspi.stop

pri getfstype : r
   if (brlong(pbuf+$36) == constant("F" + ("A" << 8) + ("T" << 16) + ("1" << 24)) and byte[pbuf][$3a]=="6")
     return 1
   if (brlong(pbuf+$52) == constant("F" + ("A" << 8) + ("T" << 16) + ("3" << 24)) and byte[pbuf][$56]=="2")
     return 2

pri hmount_explicit(DO, CLK, DI, CS) : r | start, sectorspercluster, reserved, rootentries, sectors
{{
'   Mount a volume.  The address passed in is passed along to the block
'   layer; see the currently used block layer for documentation.  If the
'   volume mounts, a 0 is returned, else abort is called.
}}
   pbuf := @buf             ' Initialize pbuf to point at buf
   if (pdate == 0)
      pdate := constant(((2009-1980) << 25) + (1 << 21) + (27 << 16) + (7 << 11))
   unmount
   sdspi.start_explicit(DO, CLK, DI, CS)
   lastread := -1
   dirty := 0
   sdspi.readblock(0, pbuf)
   if (getfstype > 0)
     start := 0
   else
     start := brlong(pbuf+$1c6)
     sdspi.readblock(start, pbuf)
   filesystem := getfstype
   if (filesystem == 0)
      abort(-20) ' not a fat16 or fat32 volume
   if (brword(pbuf+$0b) <> SECTORSIZE)
      abort(-21) ' bad bytes per sector
   sectorspercluster := byte[pbuf][$0d]
   if (sectorspercluster & (sectorspercluster - 1))
      abort(-22) ' bad sectors per cluster
   clustershift := 0
   repeat while (sectorspercluster > 1)
      clustershift++
      sectorspercluster >>= 1
   sectorspercluster := 1 << clustershift
   clustersize := SECTORSIZE << clustershift
   reserved := brword(pbuf+$0e)
   if (byte[pbuf][$10] <> 2)
      abort(-23) ' not two FATs
   sectors := brword(pbuf+$13)
   if (sectors == 0)
     sectors := brlong(pbuf+$20)
   fat1 := start + reserved
   if (filesystem == 2)
      rootentries := 16 << clustershift
      sectorsperfat := brlong(pbuf+$24)
      dataregion := (fat1 + 2 * sectorsperfat) - 2 * sectorspercluster
      rootdirstart := (dataregion + (brword(pbuf+$2c) << clustershift)) << SECTORSHIFT
      rootdirend := rootdirstart + (rootentries << DIRSHIFT)
      endofchain := $ffffff0
   else
      rootentries := brword(pbuf+$11)
      sectorsperfat := brword(pbuf+$16)
      rootdirstart := (fat1 + 2 * sectorsperfat) << SECTORSHIFT
      rootdirend := rootdirstart + (rootentries << DIRSHIFT)
      dataregion := 1 + ((rootdirend - 1) >> SECTORSHIFT) - 2 * sectorspercluster
      endofchain := $fff0
   currdirstart := rootdirstart
   currdirend := rootdirend
   long[@currdirstr] := "/"
   if (brword(pbuf+$1fe) <> $aa55)
     abort(-24) ' bad FAT signature
   totclusters := ((sectors - dataregion + start) >> clustershift)

pri readbytec(byteloc)
'
'   Read a byte address from the disk through the metadata buffer and
'   return a pointer to that location.
'
   readblockc(byteloc >> SECTORSHIFT)
   return @buf2 + (byteloc & constant(SECTORSIZE - 1))

pri readfat(clust)
'
'   Read a fat location and return a pointer to the location of that
'   entry.
'
   fatptr := (fat1 << SECTORSHIFT) + (clust << filesystem)
   return readbytec(fatptr)

pri followchain : r
'
'   Follow the fat chain and update the writelink.
'
   r := brclust(readfat(fclust))
   writelink := fatptr

pri nextcluster : r
'
'   Read the next cluster and return it.  Set up writelink to
'   point to the cluster we just read, for later updating.  If the
'   cluster number is bad, return a negative number.
'
   r := followchain
   if (r < 2 or r => totclusters)
      abort(-9) ' bad cluster value

pri freeclusters(clust) | bp
'
'   Free an entire cluster chain.  Used by remove and by overwrite.
'   Assumes the pointer has already been cleared/set to end of chain.
'
   repeat while (clust < endofchain)
      if (clust < 2)
         abort(-26) ' bad cluster number")
      bp := readfat(clust)
      clust := brclust(bp)
      brwclust(bp, 0)
   flushifdirty

pri datablock
'
'   Calculate the block address of the current data location.
'
   return (fclust << clustershift) + dataregion + ((floc >> SECTORSHIFT) & ((1 << clustershift) - 1))

pri uc(c)
'
'   Compute the upper case version of a character.
'
   if ("a" =< c and c =< "z")
      return c - 32
   return c

pri pflushbuf(rcnt, metadata) : r | cluststart, newcluster, count, i
'
'   Flush the current buffer, if we are open for write.  This may
'   allocate a new cluster if needed.  If metadata is true, the
'   metadata is written through to disk including any FAT cluster
'   allocations and also the file size in the directory entry.
'
   if (direntry == 0)
      abort(-27) ' not open for writing
   if (rcnt > 0) ' must *not* allocate cluster if flushing an empty buffer
      if (frem < SECTORSIZE)
         ' find a new clustercould be anywhere!  If possible, stay on the
         ' same page used for the last cluster.
         newcluster := -1
         cluststart := fclust & (!((SECTORSIZE >> filesystem) - 1))
         count := 2
         repeat
            readfat(cluststart)
            repeat i from 0 to SECTORSIZE - 1<<filesystem step 1<<filesystem
              if (buf2[i] == 0)
                if (brclust(@buf2+i) == 0)
                   newcluster := cluststart + (i >> filesystem)
                   if (newcluster => totclusters)
                      newcluster := -1
                   quit
            if (newcluster > 1)
               brwclust(@buf2+i, endofchain+$f)
               if (writelink == 0)
                  brwword(readbytec(direntry)+$1a, newcluster)
                  writelink := (direntry&(SECTORSIZE-filesystem))
                  brwlong(@buf2+writelink+$1c, floc+bufat)
                  if (filesystem == 2)
                     brwword(@buf2+writelink+$14, newcluster>>16)
               else
                  brwclust(readbytec(writelink), newcluster)
               writelink := fatptr + i
               fclust := newcluster
               frem := clustersize
               quit
            else
               cluststart += (SECTORSIZE >> filesystem)
               if (cluststart => totclusters)
                  cluststart := 0
                  count--
                  if (rcnt < 0)
                     rcnt := -ENOSPC ' No space left on device
                     quit
      if (frem => SECTORSIZE)
         sdspi.writeblock(datablock, pbuf)
         if (rcnt == SECTORSIZE) ' full buffer, clear it
            floc += rcnt
            frem -= rcnt
            bufat := 0
            bufend := rcnt
   if (rcnt < 0 or metadata) ' update metadata even if error
      readblockc(direntry >> SECTORSHIFT) ' flushes unwritten FAT too
      brwlong(@buf2+(direntry & (SECTORSIZE-filesystem))+$1c, floc+bufat)
      flushifdirty
   if (rcnt < 0)
      abort(rcnt)
   return rcnt

pri pfillbuf : r
'
'   Get some data into an empty buffer.  If no more data is available,
'   return -1.  Otherwise return the number of bytes read into the
'   buffer.
'
   if (floc => filesize)
      return -1
   if (frem == 0)
      fclust := nextcluster
      frem := (clustersize) <# (filesize - floc)
   sdspi.readblock(datablock, pbuf)
   r := SECTORSIZE
   if (floc + r => filesize)
      r := filesize - floc
   floc += r
   frem -= r
   bufat := 0
   bufend := r

pri pclose : r
{{
'   Flush and close the currently open file if any.  Also reset the
'   pointers to valid values.  If there is no error, 0 will be returned.
}}
   if (direntry)
      r := pflushbuf(bufat, 1)
   bufat := 0
   bufend := 0
   filesize := 0
   floc := 0
   frem := 0
   writelink := 0
   direntry := 0
   fclust := 0
   firstcluster := 0
   sdspi.release

pri setdate(year, month, day, hour, minute, second)
{{
'   Set the current date and time, as a long, in the format
'   required by FAT16.  Various limits are not checked.
}}
   pdate := ((year-1980) << 25) + (month << 21) + (day << 16)
   pdate += (hour << 11) + (minute << 5) + (second >> 1)

pri ConvertName(fname1, fname2) | i
   i := 0
   repeat while (i<8 and byte[fname1] and byte[fname1] <> ".")
      byte[fname2][i++] := uc(byte[fname1++])
   repeat while (i<8)
      byte[fname2][i++] := " "
   repeat while (byte[fname1] and byte[fname1] <> ".")
      fname1++
   if (byte[fname1] == ".")
      fname1++
   repeat while (i<11 and byte[fname1])
      byte[fname2][i++] := uc(byte[fname1++])
   repeat while (i < 11)
      byte[fname2][i++] := " "

pri popen(s, mode) : r | i, sentinel, dirptr, freeentry
{{
'   Close any currently open file, and open a new one with the given
'   file name and mode.  Mode can be "r" "w" "a" or "d" (delete).
'   If the file is opened successfully, 0 will be returned.  If the
'   file did not exist, and the mode was not "w" or "a", -1 will be
'   returned.  Otherwise abort will be called with a negative error
'   code.
'   Added "R" and "D" for directories.
}}
   pclose
   ConvertName(s, @padname)
   sentinel := 0
   freeentry := 0
   repeat dirptr from currdirstart to currdirend - DIRSIZE step DIRSIZE
      s := readbytec(dirptr)
      if (freeentry == 0 and (byte[s] == 0 or byte[s] == $e5))
         freeentry := dirptr
      if (byte[s] == 0)
         sentinel := dirptr
         quit
      repeat i from 0 to 10
         if (padname[i] <> byte[s][i])
            quit
      if (i == 11 and 0 == (byte[s][$0b] & $8)) '$18)) ' this always returns
         fclust := brword(s+$1a)
         if (filesystem == 2)
            fclust += brword(s+$14) << 16
         firstcluster := fclust
         filesize := brlong(s+$1c)
         if mode == "R" or mode == "D"
            ifnot (byte[s][11] & $10)
               abort(-ENOTDIR)
         else
            if (byte[s][11] & $10)
               abort(-EISDIR)
         if mode == "r" or mode == "R"
            frem := (clustersize) <# (filesize)
            direntryrd := dirptr
            return 0
         if (byte[s][11] & $c9)
            abort(-EACCES) ' no permission to write
         if mode == "d" or mode == "D"
            brwword(s, $e5)
            if (fclust)
               freeclusters(fclust)
            flushifdirty
            return 0
         if (mode == "w")
            brwword(s+$1a, 0)
            brwword(s+$14, 0)
            brwlong(s+$1c, 0)
            writelink := 0
            direntry := dirptr
            if (fclust)
               freeclusters(fclust)
            bufend := SECTORSIZE
            fclust := 0
            filesize := 0
            frem := 0
            return 0
         elseif (mode == "a")
            frem := filesize
            freeentry := clustersize
            if (fclust => endofchain)
               fclust := 0
            repeat while (frem > freeentry)
               if (fclust < 2)
                  abort(-7) ' eof repeat while following chain
               fclust := nextcluster
               frem -= freeentry
            floc := filesize & constant(!(SECTORSIZE - 1))
            bufend := SECTORSIZE
            bufat := frem & constant(SECTORSIZE - 1)
            writelink := 0
            direntry := dirptr
            if (bufat)
               sdspi.readblock(datablock, pbuf)
               frem := freeentry - (floc & (freeentry - 1))
            else
               if (fclust < 2 or frem == freeentry)
                  frem := 0
               else
                  frem := freeentry - (floc & (freeentry - 1))
            if (fclust => 2)
               followchain
            return 0
         else
            abort(-EINVAL) ' bad argument
   if (mode <> "w" and mode <> "a")
      return -ENOENT ' not found
   direntry := freeentry
   if (direntry == 0)
      abort(-2) ' no empty directory entry
   ' write (or new append): create valid directory entry
   s := readbytec(direntry)
   bytefill(s, 0, DIRSIZE)
   bytemove(s, @padname, 11)
   brwword(s+$1a, 0)
   brwword(s+$14, 0)
   i := pdate
   brwlong(s+$e, i) ' write create time and date
   brwlong(s+$16, i) ' write last modified date and time
   if (direntry == sentinel and direntry + DIRSIZE < currdirend)
      brwword(readbytec(direntry+DIRSIZE), 0)
   flushifdirty
   writelink := 0
   fclust := 0
   bufend := SECTORSIZE

pri pread(ubuf, count) : r | t
{{
'   Read count bytes into the buffer ubuf.  Returns the number of bytes
'   successfully read, or a negative number if there is an error.
'   The buffer may be as large as you want.
}}
   repeat while (count > 0)
      if (bufat => bufend)
         t := pfillbuf
         if (t =< 0)
            if (r > 0)
              return (r)
            return t
      t := (bufend - bufat) <# (count)
      if ((t | (ubuf) | bufat) & 3)
         bytemove(ubuf, pbuf+bufat, t)
      else
         longmove(ubuf, pbuf+bufat, t>>2)
      bufat += t
      r += t
      ubuf += t
      count -= t

pri pgetc | t
{{
'   Read and return a single character.  If the end of file is
'   reached, -1 will be returned.  If an error occurs, a negative
'   number will be returned.
}}
   if (bufat => bufend)
      t := pfillbuf
      if (t =< 0)
         return -1
   return (byte[pbuf][bufat++])

pri pwrite(ubuf, count) : r | t
{{
'   Write count bytes from the buffer ubuf.  Returns the number of bytes
'   successfully written, or a negative number if there is an error.
'   The buffer may be as large as you want.
}}
   repeat while (count > 0)
      if (bufat => bufend)
         pflushbuf(bufat, 0)
      t := (bufend - bufat) <# (count)
      if ((t | (ubuf) | bufat) & 3)
         bytemove(pbuf+bufat, ubuf, t)
      else
         longmove(pbuf+bufat, ubuf, t>>2)
      r += t
      bufat += t
      ubuf += t
      count -= t

pri pputc(c) : r
{{
'   Write a single character into the file open for write.  Returns
'   0 if successful, or a negative number if some error occurred.
}}
   if (bufat == SECTORSIZE)
     if (result := pflushbuf(SECTORSIZE, 0)) < 0
       return
   byte[pbuf][bufat++] := c

{{
'   Seek.
'
'   Returns -1 on failure.  Make sure to check this return code!
'
'   We only support reads right now (but writes won"t be too hard to
'   add).
}}
pri seek(pos) | delta
   if (direntry or pos < 0 or pos > filesize)
      return -1
   delta := (floc - bufend) & - clustersize
   if (pos < delta)
      fclust := firstcluster
      frem := (clustersize) <# (filesize)
      floc := 0
      bufat := 0
      bufend := 0
      delta := 0
   repeat while (pos => delta + clustersize)
      fclust := nextcluster
      floc += clustersize
      delta += clustersize
      frem := (clustersize) <# (filesize - floc)
      bufat := 0
      bufend := 0
   if (bufend == 0 or pos < floc - bufend or pos => floc - bufend + SECTORSIZE)
      ' must change buffer
      delta := floc + frem
      floc := pos & - SECTORSIZE
      frem := delta - floc
      pfillbuf
   bufat := pos & (SECTORSIZE - 1)
   return 0

pri popendir | off
{{
'   Close the currently open file, and set up the read buffer for
'   calls to nextfile.
}}
   pclose
   off := currdirstart - (dataregion << SECTORSHIFT)
   fclust := off >> (clustershift + SECTORSHIFT)
   floc := off - (fclust << (clustershift + SECTORSHIFT))
   frem := currdirend - currdirstart
   filesize := floc + frem
   return 0

PRI initrootdir
  currdirstart := rootdirstart
  currdirend := rootdirend

PRI initsubdir(fname)
  pbuf := @buf
  direntry := 0
  result := \popen(fname, "R")
  if result
    return
  currdirstart := ((fclust << clustershift) + dataregion) << SECTORSHIFT
  currdirend := currdirstart + 32768
  \pclose

PRI changedir(path) | nextptr, savepath
  if strcomp(path, @currdirstr)
    return
  savepath := path
  initrootdir ' Start from root directory
  ' Skip leading "/"
   if byte[path] == "/"
    path++
  ' Loop over sub-directory names in path
  repeat while byte[path]
    nextptr := path
    repeat while byte[nextptr]
      if byte[nextptr++] == "/"
        byte[nextptr-1] := 0
        quit
    if (result := initsubdir(path))
      initrootdir
      long[@currdirstr] := "/"
      return
    if byte[nextptr]
      byte[nextptr-1] := "/"
    path := nextptr
  bytemove(@currdirstr, savepath, strsize(savepath)+1)

{{
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
}}
